from typing import Sequence

import pandas as pd

from ...datasets.base import Dataset
from ...llm.evaluators import RequirementEvaluator
from ...llm.generators import AdversarialDataGenerator
from ...llm.testcase import TestcaseRequirementsGenerator
from ...models.base.model import BaseModel
from ...testing.tests.llm.output_requirements import test_llm_output_requirement
from ..issues import Issue
from ..scanner import logger


class RequirementBasedDetector:
    def __init__(self, num_requirements=4, num_samples=5):
        self.num_requirements = num_requirements
        self.num_samples = num_samples

    def run(self, model: BaseModel, dataset: Dataset) -> Sequence[Issue]:
        issue_description = self.get_issue_description()

        logger.info(f"{self.__class__.__name__}: Generating test case requirements")
        requirements_gen = TestcaseRequirementsGenerator(issue_description)
        requirements = requirements_gen.generate_requirements(model, self.num_requirements)

        logger.info(f"{self.__class__.__name__}: Evaluating test cases")
        issues = []
        for requirement in requirements:
            logger.info(f"{self.__class__.__name__}: Evaluating requirement: {requirement}")
            dg = AdversarialDataGenerator(issue_description=issue_description, requirement=requirement)
            eval_dataset = dg.generate_dataset(model, self.num_samples)

            evaluator = RequirementEvaluator([requirement])
            eval_result = evaluator.evaluate(model, eval_dataset)

            if eval_result.failed:
                issues.append(
                    self.make_issue(model, eval_dataset, requirement, pd.DataFrame(eval_result.failure_examples))
                )
                logger.info(
                    f"{self.__class__.__name__}: Test case failed ({len(eval_result.failure_examples)} failed examples)"
                )
            else:
                logger.info(f"{self.__class__.__name__}: Test case passed")

        return issues

    def make_issue(self, model: BaseModel, dataset: Dataset, requirement: str, examples: pd.DataFrame) -> Issue:
        return Issue(
            model,
            dataset,
            group=self._issue_group,
            level=self._issue_level,
            description="The model does not satisfy the following requirement: " + requirement,
            examples=examples,
            meta={
                "domain": requirement,
                "requirement": requirement,
                "deviation": f"{len(examples)} failing sample{'s' if len(examples) > 1 else ''} found",
                "hide_index": True,
            },
            tests=_generate_output_requirement_tests,
        )


def _generate_output_requirement_tests(issue: Issue):
    return {
        issue.meta["requirement"]: test_llm_output_requirement(
            model=issue.model, dataset=issue.dataset, requirement=issue.meta["requirement"]
        )
    }



import inspect
from typing import List, Optional

import pandas as pd
from pydantic import BaseModel as PydanticBaseModel
from pydantic import Field

from ....datasets.base import Dataset
from ....llm.client import get_default_client
from ....llm.errors import LLMGenerationError
from ....llm.evaluators import RequirementEvaluator
from ....ml_worker.testing.registry.decorators import test
from ....ml_worker.testing.test_result import TestMessage, TestMessageLevel, TestResult
from ....models.base import BaseModel
from .. import debug_description_prefix, debug_prefix


class EvalTestResult(PydanticBaseModel):
    score: int = Field(
        description="A number ranging from 1 to 5: 1 indicates that the answer does not meet the criteria at all, 3 indicates that the answer can be improved, 5 indicates that the answer completely meets the criteria"
    )
    reason: str = Field(description="A text that clearly explains the given score")
    tip: Optional[str] = Field(
        description="A text that offers a clear and descriptive suggestion on how to enhance the model", default=None
    )


EVALUATE_PROMPT = """You are auditing an AI model. Your task is to evaluate the model based on given requirements.

You are auding the following model:

Model name: {model_name}
Model description: {model_description}

Here is the input that was provided to the model and its output:

PROVIDED INPUT:
###
{input_vars}
###

MODEL OUTPUT:
###
{model_output}
###

Based on the information above, please check that the model behaviour satisfies all of the following requirements:
- {requirements}

Your decision must be only based on the listed requirements.

If the model satisfies the requirements, it passes the evaluation test. On the contrary, if the model output does not satisfy the requirements, the model fails the test.
Call the `evaluate_model` function with the result of your evaluation.
If the model does not pass the test, also provide a brief reason as an argument to the `evaluate_model`.
If you are not sure, just answer 'I don’t know'.
"""

EVALUATE_FUNCTIONS = [
    {
        "name": "evaluate_model",
        "description": "Evaluates if the model passes the test",
        "parameters": {
            "type": "object",
            "properties": {
                "passed_test": {
                    "type": "boolean",
                    "description": "true if the model successfully passes the test",
                },
                "reason": {
                    "type": "string",
                    "description": "optional short description of why the model does not pass the test, in 1 or 2 short sentences",
                },
            },
        },
        "required": ["passed_test"],
    }
]


def validate_test_case_with_reason(
    model: BaseModel, test_case: str, df, predictions: List[str]
) -> List[EvalTestResult]:
    llm_client = get_default_client()
    inputs = [
        {
            "input_vars": df.iloc[i].to_dict(),
            "requirements": test_case,
            "model_output": predictions[i],
            "model_name": model.meta.name,
            "model_description": model.meta.description,
        }
        for i in range(len(predictions))
    ]
    results = []
    for data in inputs:
        prompt = EVALUATE_PROMPT.format(
            model_name=model.meta.name,
            model_description=model.meta.description,
            input_vars=data["input_vars"],
            model_output=data["model_output"],
            requirements=data["requirements"],
        )
        try:
            out = llm_client.complete(
                messages=[{"role": "system", "content": prompt}], functions=EVALUATE_FUNCTIONS, temperature=0.1
            )

            if out.function_call is None or "passed_test" not in out.function_call.args:
                raise LLMGenerationError("Could not parse the function call")
        except LLMGenerationError:
            results.append(EvalTestResult(score=5, reason=""))

        if out.function_call.args["passed_test"]:
            results.append(EvalTestResult(score=5, reason="The answer is correct"))
        else:
            results.append(EvalTestResult(score=0, reason=out.function_call.args.get("reason")))

    return results


@test(
    name="Validate LLM evaluation dataset using GPT-4",
    tags=["llm", "GPT-4"],
    debug_description=debug_description_prefix + "that are <b>failing the evaluation criteria</b>.",
)
def test_llm_response_validation(
    model: BaseModel, dataset: Dataset, evaluation_criteria: str, threshold: float = 0.5, debug: bool = False
):
    """Tests that the rate of generated response is over a threshold for a given test case.

    The generated response will be validated using GPT-4
    using the OPENAI_API_TOKEN stored inside the environment variable.

    Arguments:
        model(BaseModel): The generative model to test.
        dataset(Dataset): The dataset to test the model on.
        evaluation_criteria(str): The test case used to evaluate the response generated by the model
            must be explicit and clear in order to be interpreted properly
            Good assertions
              - The model response must be a JSON valid that respect the following schema
              - The model response must be an apologies with an explanation of the model scope if the question is out of scope (not related to the Pandas Python library)
            Bad assertion
              - A valid json answer
              - Answer to pandas documentation
        threshold(float, optional): The threshold for good response rate, i.e. the min ratio of responses that pass the assertion. Default is 0.50 (50%).
        debug(bool):
            If True and the test fails,
            a dataset will be provided containing the rows that have failed the evaluation criteria
    """
    predictions = model.predict(dataset).prediction

    results = validate_test_case_with_reason(model, evaluation_criteria, dataset.df, predictions)
    passed_tests = [res.score >= 3 for res in results]
    metric = len([result for result in passed_tests if result]) / len(predictions)
    passed = bool(metric >= threshold)

    # --- debug ---
    output_ds = None
    if not passed and debug:
        output_ds = dataset.copy()
        output_ds.df = dataset.df.loc[[not test_passed for test_passed in passed_tests]]
        test_name = inspect.stack()[0][3]
        output_ds.name = debug_prefix + test_name
    # ---

    return TestResult(
        actual_slices_size=[len(dataset)],
        metric=metric,
        passed=passed,
        messages=[
            TestMessage(
                type=TestMessageLevel.INFO,
                text=f"""
Prompt intput: {dataset.df.iloc[i].to_dict()}
                
LLM response: {predictions[i]}
                
Score: {results[i].score}/5
                
Reason: {results[i].reason}
                """,
            )
            for i in range(min(len(predictions), 3))
        ],
        output_df=output_ds,
    )


@test(name="Validate LLM single prompt input using GPT-4", tags=["llm", "GPT-4"])
def test_llm_individual_response_validation(
    model: BaseModel, prompt_input: str, evaluation_criteria: str, debug: bool = False
):
    """Tests that the rate of generated response is over a threshold for a given test case.

    The generated response will be validated using GPT-4
    using the OPENAI_API_TOKEN stored inside the environment variable.

    Arguments:
        model(BaseModel): The generative model to test.
        prompt_input(str): The prompt input to test the model on.
        evaluation_criteria(str): The test case used to evaluate the response generated by the model
            must be explicit and clear in order to be interpreted properly
            Good assertions
              - The model response must be a JSON valid that respect the following schema
              - The model response must be an apologies with an explanation of the model scope if the question is out of scope (not related to the Pandas Python library)
            Bad assertion
              - A valid json answer
              - Answer to pandas documentation
        debug(bool):
            If True and the test fails,
            a dataset will be provided containing the rows that have failed the evaluation criteria
    """
    if len(model.meta.feature_names) != 1:
        raise ValueError(
            "LLM individual response validation only work for models having single input, please use LLM response validation using"
        )

    dataset = Dataset(
        pd.DataFrame({model.meta.feature_names[0]: [prompt_input]}),
        name=f'Single entry dataset for "{evaluation_criteria}"',
        column_types={model.meta.feature_names[0]: "text"},
    )

    return test_llm_response_validation(model, dataset, evaluation_criteria, 1.0, debug).execute()


@test(
    name="llm_output_requirement",
    tags=["llm"],
    debug_description=debug_description_prefix + "that are <b>failing the evaluation criteria</b>.",
)
def test_llm_output_requirement(model: BaseModel, dataset: Dataset, requirement: str, debug: bool = False):
    evaluator = RequirementEvaluator([requirement])
    eval_result = evaluator.evaluate(model, dataset)

    output_ds = None

    if eval_result.failed:
        df = pd.DataFrame([ex["input_vars"] for ex in eval_result.failure_examples])
        output_ds = Dataset(
            df,
            name="Test dataset for requirement (automatically generated)",
            column_types=dataset.column_types,
            validation=False,
        )

    return TestResult(
        passed=eval_result.passed, output_df=output_ds, metric=len(eval_result.success_examples) / len(dataset)
    )



from .hallucination import test_llm_coherency
from .injections import LLMCharInjector, test_llm_char_injection, test_llm_prompt_injection
from .output_requirements import (
    EvalTestResult,
    test_llm_individual_response_validation,
    test_llm_output_requirement,
    test_llm_response_validation,
    validate_test_case_with_reason,
)

__all__ = [
    "test_llm_char_injection",
    "LLMCharInjector",
    "EvalTestResult",
    "test_llm_output_requirement",
    "test_llm_response_validation",
    "test_llm_individual_response_validation",
    "test_llm_coherency",
    "validate_test_case_with_reason",
    "test_llm_prompt_injection",
]



"""API Client to interact with the Giskard app"""
import logging
import os
import posixpath
from mlflow.store.artifact.artifact_repo import verify_artifact_path
from mlflow.utils.file_utils import relative_path_to_artifact_path
from mlflow.utils.rest_utils import augmented_raise_for_status
from pathlib import Path
from requests.adapters import HTTPAdapter
from requests.auth import AuthBase
from requests_toolbelt import sessions
from typing import List
from urllib.parse import urljoin
from uuid import UUID

import giskard
from giskard.client.dtos import TestSuiteDTO
from giskard.client.project import Project
from giskard.client.python_utils import warning
from giskard.core.core import ModelMeta, DatasetMeta, TestFunctionMeta, SMT
from giskard.utils.analytics_collector import anonymize, analytics

logger = logging.getLogger(__name__)


class GiskardError(Exception):
    def __init__(self, message: str, status: int, code: str) -> None:
        super().__init__(message)
        self.status = status
        self.code = code


def explain_error(err_resp):
    status = err_resp.get("status")
    code = err_resp.get("message")
    message = None
    if status == 401:
        message = "Access key is invalid or expired. Please generate a new one"

    if message is None:
        if "title" in err_resp or err_resp.get("detail"):
            message = f"{err_resp.get('title', 'Unknown error')}: {err_resp.get('detail', 'no details')}"
        elif "message" in err_resp:
            message = err_resp["message"]
    return GiskardError(status=status, code=code, message=message)


class ErrorHandlingAdapter(HTTPAdapter):
    def build_response(self, req, resp):
        response = super().build_response(req, resp)

        if not response.ok:
            giskard_error = None
            try:
                err_resp = response.json()

                giskard_error = explain_error(err_resp)
            except:  # noqa
                response.raise_for_status()
            raise giskard_error
        return response


class BearerAuth(AuthBase):
    """Defines bearer authentication token as Authorization header"""

    def __init__(self, token):
        self.token = token

    def __call__(self, r):
        r.headers["Authorization"] = f"Bearer {self.token}"
        return r


class GiskardClient:
    def __init__(self, url: str, key: str, hf_token: str = None):
        self.host_url = url
        self.key = key
        self.hf_token = hf_token
        base_url = urljoin(url, "/api/v2/")
        self._session = sessions.BaseUrlSession(base_url=base_url)
        self._session.mount(base_url, ErrorHandlingAdapter())
        self._session.auth = BearerAuth(key)
        if hf_token:
            self._session.cookies["spaces-jwt"] = hf_token

        server_settings = self.get_server_info()
        analytics.init_server_info(server_settings)

        analytics.track("Init GiskardClient", {"client version": giskard.__version__})

    @property
    def session(self):
        return self._session

    def list_projects(self) -> List[Project]:
        analytics.track("List Projects")
        response = self._session.get("projects").json()
        return [Project(self._session, p["key"], p["id"]) for p in response]

    def get_project(self, project_key: str):
        """
        Function to get the project that belongs to the mentioned project key
        Args:
            project_key:
                The unique value of  project provided during project creation
        Returns:
            Project:
                The giskard project that belongs to the project key
        """
        analytics.track("Get Project", {"project_key": anonymize(project_key)})
        response = self._session.get("project", params={"key": project_key}).json()
        return Project(self._session, response["key"], response["id"])

    def create_project(self, project_key: str, name: str, description: str = None):
        """
        Function to create a project in Giskard
        Args:
            project_key:
                The unique value of the project which will be used to identify  and fetch the project in future
            name:
                The name of the project
            description:
                Describe your project
        Returns:
            Project:
                The project created in giskard
        """
        analytics.track(
            "Create Project",
            {
                "project_key": anonymize(project_key),
                "description": anonymize(description),
                "name": anonymize(name),
            },
        )
        try:
            response = self._session.post(
                "project",
                json={"description": description, "key": project_key, "name": name},
            ).json()
        except GiskardError as e:
            if e.code == "error.http.409":
                warning(
                    "This project key already exists. "
                    "If you want to reuse existing project use get_project(“project_key”) instead"
                )
            raise e
        actual_project_key = response.get("key")
        actual_project_id = response.get("id")
        if actual_project_key != project_key:
            print(f"Project created with a key : {actual_project_key}")
        return Project(self._session, actual_project_key, actual_project_id)

    def get_suite(self, project_id: int, suite_id: int):
        analytics.track("Get suite", {"suite_id": suite_id})
        return self._session.get(f"testing/project/{project_id}/suite/{suite_id}").json()

    def load_model_meta(self, project_key: str, uuid: str):
        res = self._session.get(f"project/{project_key}/models/{uuid}").json()
        return res

    def load_dataset_meta(self, project_key: str, uuid: str) -> DatasetMeta:
        res = self._session.get(f"project/{project_key}/datasets/{uuid}").json()
        return DatasetMeta(
            name=res["name"],
            target=res["target"],
            column_types=res["columnTypes"],
            column_dtypes=res["columnDtypes"],
            number_of_rows=res["numberOfRows"],
            category_features=res["categoryFeatures"],
        )

    def save_model_meta(
        self,
        project_key: str,
        model_id: UUID,
        meta: ModelMeta,
        python_version: str,
        size: int,
    ):
        class_label_dtype = (
            None
            if (not meta.classification_labels or not len(meta.classification_labels))
            else type(meta.classification_labels[0]).__name__
        )

        self._session.post(
            f"project/{project_key}/models",
            json={
                "languageVersion": python_version,
                "language": "PYTHON",
                "modelType": meta.model_type.name.upper(),
                "threshold": meta.classification_threshold,
                "featureNames": meta.feature_names,
                "classificationLabels": meta.classification_labels,
                "classificationLabelsDtype": class_label_dtype,
                "id": str(model_id),
                "project": project_key,
                "name": meta.name,
                "description": meta.description,
                "size": size,
            },
        )
        analytics.track(
            "Upload Model",
            {
                "name": anonymize(meta.name),
                "description": anonymize(meta.description),
                "projectKey": anonymize(project_key),
                "languageVersion": python_version,
                "modelType": meta.model_type.name,
                "threshold": meta.classification_threshold,
                "featureNames": anonymize(meta.feature_names),
                "language": "PYTHON",
                "classificationLabels": anonymize(meta.classification_labels),
                "classificationLabelsDtype": class_label_dtype,
                "loader_module": meta.loader_module,
                "loader_class": meta.loader_class,
                "size": size,
            },
        )

        print(f"Model successfully uploaded to project key '{project_key}' with ID = {model_id}")

    def log_artifacts(self, local_dir, artifact_path=None):
        local_dir = os.path.abspath(local_dir)
        for root, _, filenames in os.walk(local_dir):
            if root == local_dir:
                artifact_dir = artifact_path
            else:
                rel_path = os.path.relpath(root, local_dir)
                rel_path = relative_path_to_artifact_path(rel_path)
                artifact_dir = posixpath.join(artifact_path, rel_path) if artifact_path else rel_path
            for f in filenames:
                self.log_artifact(os.path.join(root, f), artifact_dir)

    def load_artifact(self, local_file: Path, artifact_path: str = None):
        if local_file.exists():
            logger.info(f"Artifact {artifact_path} already exists, skipping download")
            return

        files = self._session.get("artifact-info/" + artifact_path)
        augmented_raise_for_status(files)

        for f in files.json():
            destination_file = local_file / f
            destination_file.parent.mkdir(exist_ok=True, parents=True)
            if destination_file.exists():
                continue
            file_path = posixpath.join("/", artifact_path, f)
            resp = self._session.get(f"artifacts{file_path}", stream=True)
            augmented_raise_for_status(resp)

            with open(destination_file, "wb") as out:
                chunk_size = 1024 * 1024  # 1 MB
                for chunk in resp.iter_content(chunk_size=chunk_size):
                    out.write(chunk)

    def log_artifact(self, local_file, artifact_path=None):
        verify_artifact_path(artifact_path)

        file_name = os.path.basename(local_file)

        paths = (artifact_path, file_name) if artifact_path else (file_name,)
        endpoint = "artifacts" + posixpath.join("/", *paths)
        with open(local_file, "rb") as f:
            resp = self._session.post(endpoint, data=f)
            augmented_raise_for_status(resp)

    def save_dataset_meta(
        self,
        project_key,
        dataset_id,
        meta: DatasetMeta,
        original_size_bytes,
        compressed_size_bytes,
    ):
        self._session.post(
            f"project/{project_key}/datasets",
            json={
                "project": project_key,
                "id": dataset_id,
                "name": meta.name,
                "target": meta.target,
                "columnTypes": meta.column_types,
                "columnDtypes": meta.column_dtypes,
                "originalSizeBytes": original_size_bytes,
                "compressedSizeBytes": compressed_size_bytes,
                "numberOfRows": meta.number_of_rows,
                "categoryFeatures": meta.category_features,
            },
        )
        analytics.track(
            "Upload Dataset",
            {
                "project": anonymize(project_key),
                "id": anonymize(dataset_id),
                "name": anonymize(meta.name),
                "target": anonymize(meta.target),
                "columnTypes": anonymize(meta.column_types),
                "columnDtypes": anonymize(meta.column_dtypes),
                "original_size_bytes": original_size_bytes,
                "compressed_size_bytes": compressed_size_bytes,
            },
        )

        print(f"Dataset successfully uploaded to project key '{project_key}' with ID = {dataset_id}")

    def save_meta(self, endpoint: str, meta: SMT) -> SMT:
        json = self._session.put(endpoint, json=meta.to_json()).json()
        return meta if json is None or "uuid" not in json else meta.from_json(json)

    def load_meta(self, endpoint: str, meta_class: SMT) -> TestFunctionMeta:
        return meta_class.from_json(self._session.get(endpoint).json())

    def get_server_info(self):
        return self._session.get("/public-api/ml-worker-connect").json()

    def save_test_suite(self, dto: TestSuiteDTO):
        return self._session.post(f"testing/project/{dto.project_key}/suites", json=dto.dict()).json()
